# Hooks
from opus_instrumenta.hooks.kvs_hook import spec_variable_key_value_store_resolver

# Task Processors
from opus_instrumenta.task_processors.cli_input_prompt_v1 import CliInputPrompt
from opus_instrumenta.task_processors.shell_script_v1 import ShellScript
from opus_instrumenta.task_processors.web_download_file import WebDownloadFile

# General imports
from magnum_opus.operarius import Hook, Hooks, TaskProcessor, Tasks, LoggerWrapper, KeyValueStore, StatePersistence, TaskLifecycleStage, TaskLifecycleStages


def build_task_lifecycle_stages(task_lifecycle_stages_list: list)->TaskLifecycleStages:
    task_life_cycle_stages = TaskLifecycleStages(init_default_stages=False)
    stage: TaskLifecycleStage
    for stage in task_lifecycle_stages_list:
        if isinstance(stage, TaskLifecycleStage):
            task_life_cycle_stages.register_lifecycle_stage(task_life_cycle_stage=stage)
    return task_life_cycle_stages


STANDARD_HOOKS = {
    'spec_variable_key_value_store_resolver': {
        'Function': spec_variable_key_value_store_resolver,
        'Commands': ['ALL'],
        'Contexts': ['ALL'],
        'TaskLifeCycleStages': build_task_lifecycle_stages(task_lifecycle_stages_list=[TaskLifecycleStage.TASK_PRE_PROCESSING_START,],)
    },
}


ALL_TASK_PROCESSORS = [
    CliInputPrompt(),
    ShellScript(),
    WebDownloadFile(),
]


def build_hooks(selected_hooks: dict=STANDARD_HOOKS)->Hooks:
    hooks = Hooks()
    hook: Hook
    for hook_name, hook_config in selected_hooks.items():
        hook = Hook(
            name=hook_name,
            commands=hook_config['Commands'],
            contexts=hook_config['Contexts'],
            task_life_cycle_stages=hook_config['TaskLifeCycleStages'],
            function_impl=hook_config['Function']
        )
        hooks.register_hook(hook=hook)
    return hooks


def build_task_processors(selected_task_processors: list=ALL_TASK_PROCESSORS, logger: LoggerWrapper=LoggerWrapper())->list:
    task_processors = list()
    task_processor: TaskProcessor
    for task_processor in selected_task_processors:
        task_processor.logger = logger
        task_processors.append(task_processor)
    return task_processors


def build_tasks(
    logger: LoggerWrapper=LoggerWrapper(),
    key_value_store: KeyValueStore=KeyValueStore(),
    state_persistence: StatePersistence=StatePersistence(),
    selected_hooks: list=STANDARD_HOOKS,
    selected_task_processors: list=ALL_TASK_PROCESSORS
)->Tasks:
    tasks = Tasks(
        logger=logger,
        key_value_store=key_value_store,
        hooks=build_hooks(selected_hooks=selected_hooks),
        state_persistence=state_persistence
    )
    for task_processor in build_task_processors(selected_task_processors=selected_task_processors, logger=logger):
        tasks.register_task_processor(processor=task_processor)
    return tasks




import sys
import os
import copy
from inspect import stack
from itertools import permutations

sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../src")
print('sys.path={}'.format(sys.path))

import unittest

from opus_instrumenta.task_processors.shell_script_v1 import ShellScript
from magnum_opus.operarius import LoggerWrapper, Task, Tasks, Identifier, Identifiers, IdentifierContext, IdentifierContexts, TaskProcessor, KeyValueStore, build_command_identifier

running_path = os.getcwd()
print('Current Working Path: {}'.format(running_path))


class TestLogger(LoggerWrapper):

    def __init__(self):
        super().__init__()
        self.info_lines = list()
        self.warn_lines = list()
        self.debug_lines = list()
        self.critical_lines = list()
        self.error_lines = list()
        self.all_lines_in_sequence = list()

    def info(self, message: str):
        self.info_lines.append('[LOG] INFO: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.info_lines[-1])
        )

    def warn(self, message: str):
        self.warn_lines.append('[LOG] WARNING: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.warn_lines[-1])
        )

    def warning(self, message: str):
        self.warn_lines.append('[LOG] WARNING: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.warn_lines[-1])
        )

    def debug(self, message: str):
        self.debug_lines.append('[LOG] DEBUG: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.debug_lines[-1])
        )

    def critical(self, message: str):
        self.critical_lines.append('[LOG] CRITICAL: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.critical_lines[-1])
        )

    def error(self, message: str):
        self.error_lines.append('[LOG] ERROR: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.error_lines[-1])
        )

    def reset(self):
        self.info_lines = list()
        self.warn_lines = list()
        self.debug_lines = list()
        self.critical_lines = list()
        self.error_lines = list()


def print_logger_lines(logger:LoggerWrapper):
    for line in logger.all_lines_in_sequence:
        print(line)


def dump_key_value_store(test_class_name: str, test_method_name: str, key_value_store: KeyValueStore):
    try:
        print('\n\n-------------------------------------------------------------------------------')
        print('\t\tTest Class  : {}'.format(test_class_name))
        print('\t\tTest Method : {}'.format(test_method_name))
        print('\n-------------------------------------------------------------------------------')

        # First get the max key length:
        max_key_len = 0
        for key,val in key_value_store.store.items():
            if len(key) > max_key_len:
                max_key_len = len(key)

        for key,val in key_value_store.store.items():
            final_key = '{}'.format(key)
            spaces_qty = max_key_len - len(final_key) + 1
            spaces = ' '*spaces_qty
            final_key = '{}{}: '.format(final_key, spaces)
            print('{}{}\n'.format(final_key, val))

        print('\n_______________________________________________________________________________')
    except:
        pass


def validate_order(must_be_before_input_task_name: str, input_task_name: str, list_of_tasks: list)->bool:
    must_be_before_input_task_name_pos = list_of_tasks.index(must_be_before_input_task_name)
    input_task_name_pos = list_of_tasks.index(input_task_name)
    return must_be_before_input_task_name_pos < input_task_name_pos


class TestScenariosInLine(unittest.TestCase):    # pragma: no cover

    def setUp(self) -> None:
        print()
        print('-'*80)
        self.logger = TestLogger()
        return super().setUp()

    def tearDown(self):
        print_logger_lines(logger=self.logger)
        self.logger = None
        return super().tearDown()

    def test_order_expected_to_work_01(self):
        shell_script = ShellScript(logger=self.logger)
        t_1 = Task(
            kind='ShellScript',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "t_1"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ],
                "dependencies": [
                    {
                        "identifierType": "ManifestName",
                        "identifiers": [
                            { "key": "t_2" },
                            { "key": "t_3" },
                        ]
                    }
                ]
            },
            spec={
                'source': {
                    'type': 'inline',
                    'value': 'echo "r_1"'
                },
                'convertOutputToText': True,
                'stripNewline': True,
                'convertRepeatingSpaces': True,
                'stripLeadingTrailingSpaces': True
            },
            logger=self.logger
        )
        t_2 = Task(
            kind='ShellScript',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "t_2"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ],
                "dependencies": [
                    {
                        "identifierType": "ManifestName",
                        "identifiers": [
                            { "key": "t_3" },
                        ]
                    }
                ]
            },
            spec={
                'source': {
                    'type': 'inline',
                    'value': 'echo "r_2"'
                },
                'convertOutputToText': True,
                'stripNewline': True,
                'convertRepeatingSpaces': True,
                'stripLeadingTrailingSpaces': True
            },
            logger=self.logger
        )
        t_3 = Task(
            kind='ShellScript',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "t_3"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ]
            },
            spec={
                'source': {
                    'type': 'inline',
                    'value': 'echo "r_3"'
                },
                'convertOutputToText': True,
                'stripNewline': True,
                'convertRepeatingSpaces': True,
                'stripLeadingTrailingSpaces': True
            },
            logger=self.logger
        )
        t_4 = Task(
            kind='ShellScript',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "t_4"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ],
                "dependencies": [
                    {
                        "identifierType": "ManifestName",
                        "identifiers": [
                            { "key": "t_1" },
                            { "key": "t_2" },
                        ]
                    }
                ]
            },
            spec={
                'source': {
                    'type': 'inline',
                    'value': 'echo "r_4"'
                },
                'convertOutputToText': True,
                'stripNewline': True,
                'convertRepeatingSpaces': True,
                'stripLeadingTrailingSpaces': True
            },
            logger=self.logger
        )

        tasks_to_process = [t_1,t_2,t_3,t_4]
        permutations_of_tasks = list(permutations(tasks_to_process))

        processing_target_identifier = build_command_identifier(command='test', context='test')
        set_nr = 0
        for permutation_set in permutations_of_tasks:
            set_nr += 1
            print('Set #{}'.format(set_nr))
            tasks = Tasks(logger=self.logger)
            tasks.register_task_processor(processor=shell_script)
            task: Task
            added_task_order = list()
            for task in permutation_set:
                print('   Adding task "{}"'.format(task.task_id))
                added_task_order.append(task.task_id)
                tasks.add_task(task=task)
            calculated_task_order = tasks.calculate_current_task_order(processing_target_identifier=processing_target_identifier)
            print('   calculated_task_order={}'.format(calculated_task_order))
            tasks = None
            self.assertTrue(validate_order(must_be_before_input_task_name='t_2', input_task_name='t_1', list_of_tasks=calculated_task_order), 'Set #{} - Expected t_2 to be before t_1: added_task_order={}   calculated_task_order={}'.format(set_nr, added_task_order, calculated_task_order))
            self.assertTrue(validate_order(must_be_before_input_task_name='t_3', input_task_name='t_1', list_of_tasks=calculated_task_order), 'Set #{} - Expected t_3 to be before t_1: added_task_order={}   calculated_task_order={}'.format(set_nr, added_task_order, calculated_task_order))
            self.assertTrue(validate_order(must_be_before_input_task_name='t_3', input_task_name='t_2', list_of_tasks=calculated_task_order), 'Set #{} - Expected t_3 to be before t_2: added_task_order={}   calculated_task_order={}'.format(set_nr, added_task_order, calculated_task_order))
            self.assertTrue(validate_order(must_be_before_input_task_name='t_1', input_task_name='t_4', list_of_tasks=calculated_task_order), 'Set #{} - Expected t_1 to be before t_4: added_task_order={}   calculated_task_order={}'.format(set_nr, added_task_order, calculated_task_order))
            self.assertTrue(validate_order(must_be_before_input_task_name='t_2', input_task_name='t_4', list_of_tasks=calculated_task_order), 'Set #{} - Expected t_2 to be before t_4: added_task_order={}   calculated_task_order={}'.format(set_nr, added_task_order, calculated_task_order))

    def test_dependant_tasks_processing_with_verified_results_01(self):
        shell_script = ShellScript(logger=self.logger)
        t_1 = Task(
            kind='ShellScript',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "t_1"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ],
                "dependencies": [
                    {
                        "identifierType": "ManifestName",
                        "identifiers": [
                            { "key": "t_2" },
                            { "key": "t_3" },
                        ]
                    }
                ]
            },
            spec={
                'source': {
                    'type': 'inline',
                    'value': 'echo "r_1"'
                },
                'convertOutputToText': True,
                'stripNewline': True,
                'convertRepeatingSpaces': True,
                'stripLeadingTrailingSpaces': True
            },
            logger=self.logger
        )
        t_2 = Task(
            kind='ShellScript',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "t_2"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ],
                "dependencies": [
                    {
                        "identifierType": "ManifestName",
                        "identifiers": [
                            { "key": "t_3" },
                        ]
                    }
                ]
            },
            spec={
                'source': {
                    'type': 'inline',
                    'value': 'echo "r_2"'
                },
                'convertOutputToText': True,
                'stripNewline': True,
                'convertRepeatingSpaces': True,
                'stripLeadingTrailingSpaces': True
            },
            logger=self.logger
        )
        t_3 = Task(
            kind='ShellScript',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "t_3"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ]
            },
            spec={
                'source': {
                    'type': 'inline',
                    'value': 'echo "r_3"'
                },
                'convertOutputToText': True,
                'stripNewline': True,
                'convertRepeatingSpaces': True,
                'stripLeadingTrailingSpaces': True
            },
            logger=self.logger
        )
        t_4 = Task(
            kind='ShellScript',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "t_4"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ],
                "dependencies": [
                    {
                        "identifierType": "ManifestName",
                        "identifiers": [
                            { "key": "t_1" },
                            { "key": "t_2" },
                        ]
                    }
                ]
            },
            spec={
                'source': {
                    'type': 'inline',
                    'value': 'echo "r_4"'
                },
                'convertOutputToText': True,
                'stripNewline': True,
                'convertRepeatingSpaces': True,
                'stripLeadingTrailingSpaces': True
            },
            logger=self.logger
        )

        tasks_to_process = [t_1,t_2,t_3,t_4]
        tasks = Tasks(logger=self.logger)
        tasks.register_task_processor(processor=shell_script)
        task: Task
        for task in tasks_to_process:
            print('   Adding task "{}"'.format(task.task_id))
            tasks.add_task(task=task)
        tasks.process_context(command='apply', context='unittest')
        dump_key_value_store(test_class_name=self.__class__.__name__, test_method_name=stack()[0][3], key_value_store=tasks.key_value_store)
        for task_id in ('t_1','t_2','t_3','t_4',):

            self.assertTrue('PROCESSING_TASK:{}:apply:unittest'.format(task_id) in tasks.key_value_store.store)
            self.assertIsInstance(tasks.key_value_store.store['PROCESSING_TASK:{}:apply:unittest'.format(task_id)], int)
            self.assertEqual(tasks.key_value_store.store['PROCESSING_TASK:{}:apply:unittest'.format(task_id)], 2)

            self.assertTrue('ShellScript:{}:apply:unittest:processing:result:EXIT_CODE'.format(task_id) in tasks.key_value_store.store)
            self.assertIsInstance(tasks.key_value_store.store['ShellScript:{}:apply:unittest:processing:result:EXIT_CODE'.format(task_id)], int)
            self.assertEqual(tasks.key_value_store.store['ShellScript:{}:apply:unittest:processing:result:EXIT_CODE'.format(task_id)], 0)

            self.assertTrue('ShellScript:{}:apply:unittest:processing:result:STDERR'.format(task_id) in tasks.key_value_store.store)
            self.assertIsInstance(tasks.key_value_store.store['ShellScript:{}:apply:unittest:processing:result:STDERR'.format(task_id)], str)
            self.assertEqual(tasks.key_value_store.store['ShellScript:{}:apply:unittest:processing:result:STDERR'.format(task_id)], '')

            result = task_id.replace('t', 'r')
            self.assertTrue('ShellScript:{}:apply:unittest:processing:result:STDOUT'.format(task_id) in tasks.key_value_store.store)
            self.assertIsInstance(tasks.key_value_store.store['ShellScript:{}:apply:unittest:processing:result:STDOUT'.format(task_id)], str)
            self.assertEqual(tasks.key_value_store.store['ShellScript:{}:apply:unittest:processing:result:STDOUT'.format(task_id)], result)


if __name__ == '__main__':
    unittest.main()



import copy
import traceback
from pathlib import Path
import subprocess
import tempfile
import chardet
import os
import json
from magnum_opus.operarius import LoggerWrapper, TaskProcessor, KeyValueStore, Task, StatePersistence


class ShellScript(TaskProcessor):
    """The `ShellScript` task processor executes a shell script based on the provided spec.

    By default, any `ShellScript` task will be processed, regardless of `command` and `context`. To exclude a 
    `ShellScript` task from being processed in specific cases, use the appropriate contextual identifiers, for example:

    ```python
    metadata = {
        "contextualIdentifiers": [
            {
                "type": 'ExecutionScope',
                "key": 'INCLUDE',               # Only consider processing this task if the supplied processing context
                "contexts": [                   # is one of the listed environments
                    {
                        "type": "Environment",
                        "names": [
                            "sandbox",
                            "test",
                            "prod"
                        ]
                    }
                ]
            },
            {
                "type": 'ExecutionScope',
                "key": 'EXCLUDE',               # Specifically exclude this task from being processed during "delete"
                "contexts": [                   # commands
                    {
                        "type": "Command",
                        "names": [
                            "delete"
                        ]
                    }
                ]
            }
        ]
    }
    task = Task(kind='ShellScript', version='v1', spec=..., metadata=metadata, logger=...)
    ```

    Attributes:
        logger: An implementation of the `LoggerWrapper` class
        kind: The kind. Any `Task` with the same kind (and matching version) may be processed with this task processor.
        versions: A list of supported versions that this task processor can process
        supported_commands: A list of supported commands that this task processor can process on matching tasks.
    """

    def __init__(self, kind: str='ShellScript', kind_versions: list=['v1',], supported_commands: list = list(), logger: LoggerWrapper = LoggerWrapper()):
        self.spec = dict()
        self.metadata = dict()
        super().__init__(kind, kind_versions, supported_commands, logger)

    def _id_source(self, log_header: str='')->str:
        source = 'inline'
        if 'source' in self.spec:
            if 'type' in self.spec['source']:
                if self.spec['source']['type'] in ('inLine', 'filePath',):
                    source = self.spec['source']['type']
        return source

    def _load_source_from_spec(self, log_header: str='')->str:
        source = 'exit 0'
        if 'source' in self.spec:
            if 'value' in self.spec['source']:
                source = self.spec['source']['value']
        return source

    def _load_source_from_file(self, log_header: str='')->str:
        source = 'exit 0'
        if 'source' in self.spec:
            if 'value' in self.spec['source']:
                try:
                    self.log(message='   Loading script source from file "{}"'.format(self.spec['source']['value']), level='info', build_log_message_header=False , header=log_header)
                    with open(self.spec['source']['value'], 'r') as f:
                        source = f.read()
                except:
                    self.log(message='   EXCEPTION: {}'.format(traceback.format_exc()), level='error', build_log_message_header=False , header=log_header)
        return source

    def _get_work_dir(self, log_header: str='')->str:
        work_dir = tempfile.gettempdir()
        if 'workdir' in self.spec:
            if 'path' in self.spec['workdir']:
                work_dir = self.spec['workdir']['path']
        self.log(message='   Work directory set to "{}"'.format(work_dir), level='info', build_log_message_header=False , header=log_header)
        return work_dir

    def _del_file(self, file: str, log_header: str=''):
        try:
            os.unlink(file)
        except:
            pass

    def _create_work_file(self, source:str, log_header: str='', task_id: str='not-set')->str:
        work_file = '{}{}{}'.format(
            self._get_work_dir(log_header=log_header),
            os.sep,
            task_id
        )
        self.log(message='   Writing source code to file "{}"'.format(work_file), level='info', build_log_message_header=False , header=log_header)
        self._del_file(file=work_file)
        try:
            with open(work_file, 'w') as f:
                f.write(source)
            self.log(message='      DONE', level='info', build_log_message_header=False , header=log_header)
        except:
            self.log(message='   EXCEPTION in _create_work_file(): {}'.format(traceback.format_exc()), level='error', build_log_message_header=False , header=log_header)
        return work_file

    def __detect_encoding(self, input_str: str)->str:
        encoding = None
        try:
            encoding = chardet.detect(input_str)['encoding']
        except:
            pass
        return encoding

    def process_task(self, task: Task, command: str, context: str='default', key_value_store: KeyValueStore=KeyValueStore(), state_persistence: StatePersistence=StatePersistence())->KeyValueStore:
        """This `Task` runs a shell command and saves the shell exitcode, STDOUT and STDERR values to the 
        `KeyValueStore`
        
        Regardless of command and context, the specified shell script will be run, unless specifically excluded.

        # Spec fields

        Root levels spec fields

        | Field                        | Type     | Required | In Versions | Description                                                                                                                                                                                                                                     |
        | ---------------------------- | :------: | :------: | :---------: | ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |
        | `shellInterpreter`           |  str     |    No    |     v1      | The shell interpreter to select in the shabang line. Supported values: `sh`, `zsh`, `perl`, `python` and `bash`                                                                                                                                 |
        | `source`                     |  dict    |    Yes   |     v1      | Defines the script source                                                                                                                                                                                                                       |
        | `workDir`                    |  dict    |    No    |     v1      | Defines work directory attributes..                                                                                                                                                                                                             |
        | `convertOutputToText`        |  bool    |    No    |     v1      | Normally the STDOUT and STDERR will be binary encoded. Setting this value to true will convert those values to a normal string. Default=False                                                                                                   |
        | `stripNewline`               |  bool    |    No    |     v1      | Output may include newline or other line break characters. Setting this value to true will remove newline characters. Default=False                                                                                                             |
        | `convertRepeatingSpaces`     |  bool    |    No    |     v1      | Output may contain more than one repeating space or tab characters. Setting this value to true will replace these with a single space. Default=False                                                                                            |
        | `stripLeadingTrailingSpaces` |  bool    |    No    |     v1      | Output may contain more than one repeating space or tab characters. Setting this value to true will replace these with a single space. Default=False                                                                                            |
        | `raiseExceptionOnError`      |  bool    |    No    |     v1      | Default value is `False`. If set to `True`, and shell processing exit code other that `0` will force an exception to be raised.                                                                                                                 |

        ## Fields for `source`

        | Field                        | Type     | Required | In Versions | Description                                                                                                                                                                                                                                     |
        | ---------------------------- | :------: | :------: | :---------: | ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |
        | `type`                       |  str     |    No    |     v1      | Select the source type, which can be either `filePath` that points to an existing script file on the local file system, or `inLine` with the script source defined in the `spec.source.value` field                                             |
        | `value`                      |  str     |    No    |     v1      | If `spec.source.type` has a value of `inLine` then the value here will be assumed to be the script content of that type. if `spec.source.type` has a value of `filePath` then this value must point to an existing file on the local filesystem |

        ## Fields for `workDir`

        | Field                        | Type     | Required | In Versions | Description                                                                                                                                                                                                                                     |
        | ---------------------------- | :------: | :------: | :---------: | ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |
        | `path`                       |  str     |    No    |     v1      | An optional path to a working directory. The `ShellScript` will create temporary files (if needed) in this directory and execute them from here.                                                                                                |        

        Args:
            task: The `Task` of kind `ShellScript` version `v1` to process
            command: The command is ignored for ShellScript processing - any task of this kind will ALWAYS be processed regardless of the command.
            context: The context is ignored for ShellScript processing - any task of this kind will ALWAYS be processed regardless of the context.
            key_value_store: An instance of the `KeyValueStore`. If none is supplied, a new instance will be created.
            state_persistence: An implementation of `StatePersistence` that the task processor can use to retrieve previous copies of the `Task` manifest in order to determine the actions to be performed.

        Returns:
            An updated `KeyValueStore`.

            * Output from STDOUT will be saved under the key: `task.kind:task.task_id:command:context:processing:result:STDOUT`
            * Output from STDERR will be saved under the key: `task.kind:task.task_id:command:context:processing:result:STDERR`
            * Script exit code will be saved under the key: `task.kind:task.task_id:command:context:processing:result:EXIT_CODE`

        Raises:
            Exception: As determined by the processing logic.
        """
        
        result_stdout = ''
        result_stderr = ''
        result_exit_code = 0
        new_key_value_store = KeyValueStore()
        new_key_value_store.store = copy.deepcopy(key_value_store.store)
        log_header = self.format_log_header(task=task, command=command, context=context)
        self.log(message='PROCESSING START', build_log_message_header=False, level='info', header=log_header)
        self.log(message='   spec: {}'.format(json.dumps(self.spec)), build_log_message_header=False, level='debug', header=log_header)
        if '{}:{}:{}:{}:processing:result:EXIT_CODE'.format(task.kind,task.task_id,command,context) in key_value_store.store is True:
            self.log(message='The task have already been processed and will now be ignored. The KeyValueStore will be returned unmodified.', build_log_message_header=False, level='warning', header=log_header)
            return new_key_value_store
        
        task_processing_exception_raised = False
        task_processing_exception_formatted_stacktrace = ''
        try:
            self.spec = copy.deepcopy(task.spec)
            self.metadata = copy.deepcopy(task.metadata)
            self.log(message='spec={}'.format(json.dumps(self.spec)), build_log_message_header=False, level='debug', header=log_header)

            ###
            ### PREP SOURCE FILE
            ###
            script_source = 'exit 0'
            if self._id_source() == 'inline':
                shabang = '#!/bin/sh'
                if 'shellinterpreter' in self.spec:
                    shabang = self.spec['shellinterpreter']
                    script_source = '#!/usr/bin/env {}\n\n{}'.format(
                        shabang,
                        self._load_source_from_spec()
                    )
                else:
                    script_source = '{}\n\n{}'.format(
                        shabang,
                        self._load_source_from_spec()
                    )
            else:
                script_source = self._load_source_from_file()
            self.log(message='script_source:\n--------------------\n{}\n--------------------'.format(script_source), build_log_message_header=False, level='debug', header=log_header)
            work_file = self._create_work_file(source=script_source, task_id=task.task_id, log_header=log_header)

            ###
            ### EXECUTE
            ###
            result = None
            os.chmod(work_file, 0o700)
            result = subprocess.run('{}'.format(work_file), check=True, capture_output=True)   # Returns CompletedProcess

            ###
            ### STORE VALUES
            ###
            if result is not None:
                result_stdout = result.stdout
                result_exit_code = result.returncode
                self.log(message='   Storing Variables', build_log_message_header=False, level='info', header=log_header)                
                value_stdout_encoding = self.__detect_encoding(input_str=result.stdout)
                value_stderr_encoding = self.__detect_encoding(input_str=result.stdout)
                result_stderr = result.stderr

                if 'convertoutputtotext' in self.spec:
                    self.log(message='      Processing "convertOutputToText"', build_log_message_header=False, level='debug', header=log_header)   
                    if self.spec['convertoutputtotext'] is True:
                        if value_stdout_encoding is not None:
                            result_stdout = result_stdout.decode(value_stdout_encoding)
                        if value_stderr_encoding is not None:
                            result_stderr = result_stderr.decode(value_stderr_encoding)

                if 'stripnewline' in self.spec:
                    self.log(message='      Processing "stripNewline"', build_log_message_header=False, level='debug', header=log_header)   
                    if self.spec['stripnewline'] is True:
                        try:
                            if result_stdout is not None:
                                result_stdout = result_stdout.replace('\n', '')
                                result_stdout = result_stdout.replace('\r', '')
                            if result_stderr is not None:
                                result_stderr = result_stderr.replace('\n', '')
                                result_stderr = result_stderr.replace('\r', '')
                        except:
                            traceback.print_exc()
                            self.log(message='Could not remove newline characters after "StripNewline" setting was set to True', build_log_message_header=False, level='warning', header=log_header)

                if 'convertrepeatingspaces' in self.spec:
                    self.log(message='      Processing "convertRepeatingSpaces"', build_log_message_header=False, level='debug', header=log_header)
                    if self.spec['convertrepeatingspaces'] is True:
                        try:
                            if result_stdout is not None:
                                result_stdout = ' '.join(result_stdout.split())
                            if result_stderr is not None:
                                result_stderr = ' '.join(result_stderr.split())
                        except:
                            traceback.print_exc()
                            self.log(message='Could not remove repeating whitespace characters after "ConvertRepeatingSpaces" setting was set to True', build_log_message_header=False, level='warning', header=log_header)

                if 'stripleadingtrailingspaces' in self.spec:
                    self.log(message='      Processing "stripLeadingTrailingSpaces"', build_log_message_header=False, level='debug', header=log_header)
                    if self.spec['stripleadingtrailingspaces'] is True:
                        try:
                            if result_stdout is not None:
                                result_stdout = result_stdout.strip()
                            if result_stderr is not None:
                                result_stderr = result_stderr.strip()
                        except:
                            traceback.print_exc()
                            self.log(message='Could not remove repeating whitespace characters after "ConvertRepeatingSpaces" setting was set to True', build_log_message_header=False, level='warning', header=log_header)

        except:
            task_processing_exception_formatted_stacktrace = traceback.format_exc()
            self.log(message='EXCEPTION: {}'.format(task_processing_exception_formatted_stacktrace), build_log_message_header=False, level='error', header=log_header)
            task_processing_exception_raised = True

        self.spec = dict()
        self.metadata = dict()

        if task_processing_exception_raised is True or result_exit_code != 0:
            if 'raiseExceptionOnError' in task.spec:
                if isinstance(task.spec['raiseExceptionOnError'], bool):
                    if task.spec['raiseExceptionOnError'] is True:
                        raise Exception('{} - Task Processing failed Stacktrace was logged.'.format(log_header))
            if len(task_processing_exception_formatted_stacktrace) > 0:
                if len(result_stderr) != 0:
                    result_stderr = '{}\n\n'.format(result_stderr)
                result_stderr = '{}Processing Exception Stacktrace:\n-------------------------------\n{}\n-------------------------------'.format(
                    result_stderr,
                    task_processing_exception_formatted_stacktrace
                )
        
        self.log(message='      Storing Exit Code', build_log_message_header=False, level='info', header=log_header)
        new_key_value_store.save(key='{}:{}:{}:{}:processing:result:EXIT_CODE'.format(task.kind, task.task_id, command, context), value=result_exit_code)

        self.log(message='      Storing STDERR', build_log_message_header=False, level='info', header=log_header)
        new_key_value_store.save(key='{}:{}:{}:{}:processing:result:STDERR'.format(task.kind, task.task_id, command, context), value=result_stderr)

        self.log(message='      Storing STDOUT', build_log_message_header=False, level='info', header=log_header)
        new_key_value_store.save(key='{}:{}:{}:{}:processing:result:STDOUT'.format(task.kind, task.task_id, command, context), value=result_stdout)

        self.log(message='DONE', build_log_message_header=False, level='info', header=log_header)
        return new_key_value_store




import sys
import os
import copy
from inspect import stack

sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../src")
print('sys.path={}'.format(sys.path))

import unittest

from opus_instrumenta.task_processors.cli_input_prompt_v1 import CliInputPrompt
from magnum_opus.operarius import LoggerWrapper, Task, Tasks, Identifier, Identifiers, IdentifierContext, IdentifierContexts, TaskProcessor, KeyValueStore

running_path = os.getcwd()
print('Current Working Path: {}'.format(running_path))


class TestLogger(LoggerWrapper):

    def __init__(self):
        super().__init__()
        self.info_lines = list()
        self.warn_lines = list()
        self.debug_lines = list()
        self.critical_lines = list()
        self.error_lines = list()
        self.all_lines_in_sequence = list()

    def info(self, message: str):
        self.info_lines.append('[LOG] INFO: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.info_lines[-1])
        )

    def warn(self, message: str):
        self.warn_lines.append('[LOG] WARNING: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.warn_lines[-1])
        )

    def warning(self, message: str):
        self.warn_lines.append('[LOG] WARNING: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.warn_lines[-1])
        )

    def debug(self, message: str):
        self.debug_lines.append('[LOG] DEBUG: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.debug_lines[-1])
        )

    def critical(self, message: str):
        self.critical_lines.append('[LOG] CRITICAL: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.critical_lines[-1])
        )

    def error(self, message: str):
        self.error_lines.append('[LOG] ERROR: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.error_lines[-1])
        )

    def reset(self):
        self.info_lines = list()
        self.warn_lines = list()
        self.debug_lines = list()
        self.critical_lines = list()
        self.error_lines = list()


def print_logger_lines(logger:LoggerWrapper):
    for line in logger.all_lines_in_sequence:
        print(line)


def dump_key_value_store(test_class_name: str, test_method_name: str, key_value_store: KeyValueStore):
    try:
        print('\n\n-------------------------------------------------------------------------------')
        print('\t\tTest Class  : {}'.format(test_class_name))
        print('\t\tTest Method : {}'.format(test_method_name))
        print('\n-------------------------------------------------------------------------------')

        # First get the max key length:
        max_key_len = 0
        for key,val in key_value_store.store.items():
            if len(key) > max_key_len:
                max_key_len = len(key)

        for key,val in key_value_store.store.items():
            final_key = '{}'.format(key)
            spaces_qty = max_key_len - len(final_key) + 1
            spaces = ' '*spaces_qty
            final_key = '{}{}: '.format(final_key, spaces)
            print('{}{}\n'.format(final_key, val))

        print('\n_______________________________________________________________________________')
    except:
        pass


class TestScenariosInLine(unittest.TestCase):    # pragma: no cover

    def setUp(self) -> None:
        print()
        print('-'*80)
        self.logger = TestLogger()
        return super().setUp()

    def tearDown(self):
        print_logger_lines(logger=self.logger)
        self.logger = None
        return super().tearDown()

    def test_echo_hello_world_01(self):
        cli_input = CliInputPrompt(logger=self.logger)
        task = Task(
            kind='CliInputPrompt',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "test1"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ]
            },
            spec={
                'promptText': 'This is a test that will return "it worked!" after 2 seconds timeout',
                'defaultValue': 'it worked!',
                'promptCharacter': '$$',
                'waitTimeoutSeconds': 2,
            },
            logger=self.logger
        )
        tasks = Tasks(logger=self.logger)
        tasks.register_task_processor(processor=cli_input)
        tasks.add_task(task=task)
        tasks.process_context(command='apply', context='unittest')
        tasks.state_persistence.persist_all_state()
        dump_key_value_store(test_class_name=self.__class__.__name__, test_method_name=stack()[0][3], key_value_store=tasks.key_value_store)
        self.assertIsNotNone(tasks.key_value_store)
        self.assertIsNotNone(tasks.key_value_store.store)
        self.assertIsInstance(tasks.key_value_store, KeyValueStore)
        self.assertIsInstance(tasks.key_value_store.store, dict)
        self.assertTrue('PROCESSING_TASK:test1:apply:unittest' in tasks.key_value_store.store)
        self.assertTrue('CliInputPrompt:test1:apply:unittest:RESULT' in tasks.key_value_store.store)
        self.assertEqual(tasks.key_value_store.store['CliInputPrompt:test1:apply:unittest:RESULT'], 'it worked!')


if __name__ == '__main__':
    unittest.main()



import os
import json
import traceback
import copy
from pathlib import Path
import requests
from requests.auth import HTTPBasicAuth

from magnum_opus.operarius import LoggerWrapper, TaskProcessor, KeyValueStore, Task, StatePersistence
from magnum_opus_adstator.file_io import get_file_size


class WebDownloadFile(TaskProcessor):

    def __init__(self, kind: str='WebDownloadFile', kind_versions: list=['v1',], supported_commands: list = list(), logger: LoggerWrapper = LoggerWrapper()):
        self.spec = dict()
        self.metadata = dict()
        super().__init__(kind, kind_versions, supported_commands, logger)

    def _get_url_content_length(self, url: str, log_header:str='')->dict:
        try:
            response = requests.head(url, allow_redirects=True)
            self.log(message='Headers: {}'.format(response.headers), level='debug')
            for header_name, header_value in response.headers.items():
                if header_name.lower() == 'content-length':
                    self.log(message='Content-Length: {}'.format(int(header_value)), build_log_message_header=False, level='info', header=log_header)
                    return int(header_value)
        except:
            self.log(message='EXCEPTION: {}'.format(traceback.format_exc()), build_log_message_header=False, level='error', header=log_header)
        # It may be impossible to get the initial length as we did not take into account proxy or authentication. In these cases assume a LARGE file
        self.log(message='Unable to determine content length from URL {}'.format(url), build_log_message_header=False, level='warning', header=log_header)
        return 999999999999

    def _build_proxy_dict(self, proxy_host: str, proxy_username: str, proxy_password: str, log_header:str='')->dict:
        proxies = dict()
        proxy_str = ''
        if proxy_host is not None:
            if isinstance(proxy_host, str):
                if proxy_host.startswith('http'):
                    proxy_str = proxy_host
                    if proxy_username is not None and proxy_password is not None:
                        if isinstance(proxy_username, str) and isinstance(proxy_password, str):
                            creds = '//{}:{}@'.format(proxy_username, proxy_password)
                            creds_logging = '//{}:{}@'.format(proxy_username, '*' * len(proxy_password))
                            final_proxy_str = '{}{}{}'.format(proxy_str.split('/')[0], creds, '/'.join(proxy_str.split('/')[2:]))
                            final_proxy_str_logging = '{}{}{}'.format(proxy_str.split('/')[0], creds_logging, '/'.join(proxy_str.split('/')[2:]))
                            self.log(message='Using proxy "{}"'.format(final_proxy_str_logging), build_log_message_header=False, level='info', header=log_header)
                            proxies['http'] = final_proxy_str
                            proxies['https'] = final_proxy_str
        return proxies

    def _build_http_basic_auth_dict(self, username: str, password: str, log_header:str='')->HTTPBasicAuth:
        auth = None
        if username is not None and password is not None:
            if len(username) > 0 and len(password) > 0:
                auth = HTTPBasicAuth(username, password)
        return auth

    def _get_data_basic_request(
        self,
        url: str,
        target_file: str,
        verify_ssl: bool,
        proxy_host: str,
        proxy_username: str,
        proxy_password: str,
        username: str,
        password: str,
        headers: dict,
        method: str,
        body: str,
        log_header:str=''
    )->str:
        self.log(message='Running Method "_get_data_basic_request()"', build_log_message_header=False, level='debug', header=log_header)
        http_status_code = None
        try:
            proxies=self._build_proxy_dict(proxy_host=proxy_host, proxy_username=proxy_username, proxy_password=proxy_password)
            auth = self._build_http_basic_auth_dict(username=username, password=password)
            r = requests.request(method=method, url=url, allow_redirects=True, verify=verify_ssl, proxies=proxies, auth=auth, headers=headers, data=body)
            if r:
                http_status_code = '{}'.format(r.status_code)
            if r.content:
                with open(target_file, 'wb') as f:
                    f.write(r.content)
        except:
            self.log(message='EXCEPTION: {}'.format(traceback.format_exc()), build_log_message_header=False, level='error', header=log_header)
        return http_status_code

    def _get_data_basic_request_stream(
        self,
        url: str,
        target_file: str,
        verify_ssl: bool,
        proxy_host: str,
        proxy_username: str,
        proxy_password: str,
        username: str,
        password: str,
        headers: dict,
        method: str,
        body: str,
        log_header:str=''
    )->bool:
        # Refer to https://stackoverflow.com/questions/16694907/download-large-file-in-python-with-requests
        self.log(message='Running Method "_get_data_basic_request_stream()"', build_log_message_header=False, level='debug', header=log_header)
        http_status_code = None
        try:
            proxies=self._build_proxy_dict(proxy_host=proxy_host, proxy_username=proxy_username, proxy_password=proxy_password)
            auth = self._build_http_basic_auth_dict(username=username, password=password)
            with requests.request(method=method, url=url, allow_redirects=True, verify=verify_ssl, proxies=proxies, auth=auth, headers=headers, stream=True, data=body) as r:
                if r:
                    http_status_code = '{}'.format(r.status_code)
                r.raise_for_status()
                with open(target_file, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)
        except:
            self.log(message='EXCEPTION: {}'.format(traceback.format_exc()), build_log_message_header=False, level='error', header=log_header)
            return None
        return http_status_code

    def _store_values(self, key_value_store: KeyValueStore, value: object, task_id: str, command: str, context: str, log_header:str='')->KeyValueStore:
        new_key_value_store = KeyValueStore()
        new_key_value_store.store = copy.deepcopy(key_value_store.store)
        final_key = '{}:{}:{}:{}:RESULT'.format(self.kind, task_id, command, context)
        self.log(message='  Storing value in key "{}"'.format(final_key), build_log_message_header=False, level='info', header=log_header)
        new_key_value_store.save(key=final_key, value=value)
        return new_key_value_store

    def delete_output_file(
        self,
        task_id: str,
        command: str,
        context: str,
        key_value_store: KeyValueStore=KeyValueStore(),
        remove_target_output_file_stored_result_on_file_deletion: bool=True,
        log_header: str=''
    )->KeyValueStore:
        """A helper function that the client can use to delete a previously downloaded artifact, or an artifact that is
        defined in the `spec` that may exist on the filesystem.

        Args:
            task_id: The task ID
            command: The command
            context: The context
            key_value_store: A copy of the current `KeyValueStore`. A new instance with updated values will be returned.
            remove_target_output_file_stored_result_on_file_deletion: A boolean indicating if the `KeyValueStore` result must also be removed when the file is deleted. The default is `True`.
            log_header: A string that is added as a leader string to all log messages. The string can be empty, but not `None`

        Returns:
            A new instance of `KeyValueStore`. 
            
            If the argument `remove_target_output_file_stored_result_on_file_deletion` was set to `True` and a previous
            download was done with a stored result, the previous result data will be removed from the `KeyValueStore`.
        """
        new_key_value_store = KeyValueStore()
        new_key_value_store.store = copy.deepcopy(key_value_store.store)
        try:
            if 'targetOutputFile' in self.spec:
                if self.spec['targetOutputFile'] is not None:
                    if isinstance(self.spec['targetOutputFile'], str):
                        os.unlink(self.spec['targetOutputFile'])            
        except:
            self.log(message='EXCEPTION: {}'.format(traceback.format_exc()), build_log_message_header=False, level='error', header=log_header)
        if remove_target_output_file_stored_result_on_file_deletion is True:
            key = 'WebDownloadFile:{}:{}:{}:RESULT'.format(task_id,command,context)
            if key in key_value_store.store is True:
                new_key_value_store.store.pop(key)
        return new_key_value_store

    def process_task(self, task: Task, command: str, context: str = 'default', key_value_store: KeyValueStore = KeyValueStore(), state_persistence: StatePersistence = StatePersistence()) -> KeyValueStore:
        """This task processor will download a file from a HTTP or HTTPS server.

        # Spec fields

        Root levels spec fields

        | Field                     | Type    | Required | In Versions | Description                                                                                                                                                                                    |
        |---------------------------|:-------:|:--------:|:-----------:|------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
        | `sourceUrl`               | string  | Yes      | v1          | The URL from where to download the file                                                                                                                                                        |
        | `targetOutputFile`        | string  | Yes      | v1          | The destination file. NOTE: The directory MUST exist. To create the directory first (if needed) consider using a ShellScript as a dependency.                                                  |
        | `skipSslVerification`     | bool    | No       | v1          | If set to true, skips SSL verification. WARNING: use with caution as this may pose a serious security risk                                                                                     |
        | `proxy`                   | dict    | No       | v1          | Proxy configuration.                                                                                                                                                                           |
        | `extraHeaders`            | list    | No       | v1          | A list of name and value items with additional headers to set for the request. Things like a Authorization header might need to be set.                                                        |
        | `method`                  | string  | No       | v1          | The HTTP method to use (default=GET)                                                                                                                                                           |
        | `body`                    | string  | No       | v1          | Some request types, like POST, requires a body with the data to send. Also remember to set additional headers like "Content Type" as required                                                  |
        | `httpBasicAuthentication  | dict    | No       | v1          | If the remote site requires basic authentication, set the username using this field                                                                                                            |
        | `successCodes`            | string  | No       | v1          | A string describing the HTTP return codes to be considered as success. Any other code besides this will be considered an error state. Default: `200-399`                                       |
        | `exceptionOnError`        | bool    | No       | v1          | If set to `True`, any HTTP return value considered to be an error will result in the processing raising an error. Setting this value to `False` will not raise an `Exception`. Default: `True` |

        ## Fields for `proxy`

        | Field                 | Type    | Required    | In Versions | Description                                                                                                                                                         |
        |-----------------------|:-------:|:-----------:|:-----------:|---------------------------------------------------------------------------------------------------------------------------------------------------------------------|
        | `host`                | string  | Conditional | v1          | If you need to pass through a proxy, set the proxy host here. Include the protocol and port, for example `http://` or `https://`. An example: `http://myproxy:3128` |
        | `basicAuthentication` | dict    | No          | v1          | Proxy basic username and password configuration                                                                                                                     |

        NOTE: If the environment variables `http_proxy`, `https_proxy`, `HTTP_PROXY` or `HTTPS_PROXY`, those values will
        be used if `host` (or the root `proxy` field) is not supplied. Setting the `host` value here will override any
        environment variables.

        ### Fields for `basicAuthentication`

        | Field      | Type    | Required | In Versions | Description                                                                                   |
        |------------|:-------:|:--------:|:-----------:|-----------------------------------------------------------------------------------------------|
        | `username` | string  | No       | v1          | If the proxy requires authentication and supports basic authentication, set the username here |
        | `password` | string  | No       | v1          | The password.                                                                                 |

        ## Fields for `httpBasicAuthentication`

        | Field      | Type    | Required | In Versions | Description                                                                                   |
        |------------|:-------:|:--------:|:-----------:|-----------------------------------------------------------------------------------------------|
        | `username` | string  | No       | v1          | If the proxy requires authentication and supports basic authentication, set the username here |
        | `password` | string  | No       | v1          | The password.                                                                                 |
 
        Args:
            task: The `Task` of kind `ShellScript` version `v1` to process
            command: The command is ignored for ShellScript processing - any task of this kind will ALWAYS be processed regardless of the command.
            context: The context is ignored for ShellScript processing - any task of this kind will ALWAYS be processed regardless of the context.
            key_value_store: An instance of the `KeyValueStore`. If none is supplied, a new instance will be created.
            state_persistence: An implementation of `StatePersistence` that the task processor can use to retrieve previous copies of the `Task` manifest in order to determine the actions to be performed.

        Returns:
            An updated `KeyValueStore`.

            * The HTTP return code will be stored in: `task.kind:task.task_id:command:context:RESULT`

        Raises:
            Exception: As determined by the processing logic.
        """
        self.spec = copy.deepcopy(task.spec)
        self.metadata = copy.deepcopy(task.metadata)
        new_key_value_store = KeyValueStore()
        new_key_value_store.store = copy.deepcopy(key_value_store.store)
        log_header = self.format_log_header(task=task, command=command, context=context)
        self.log(message='PROCESSING START', build_log_message_header=False, level='info', header=log_header)
        self.log(message='   spec: {}'.format(json.dumps(self.spec)), build_log_message_header=False, level='debug', header=log_header)
        if '{}:{}:{}:{}:RESULT'.format(task.kind,task.task_id,command,context) in key_value_store.store is True:
            self.log(message='The task have already been processed and will now be ignored. The KeyValueStore will be returned unmodified.', build_log_message_header=False, level='warning', header=log_header)
            return new_key_value_store

        url: str
        url = None
        if 'sourceurl' in self.spec:
            if self.spec['sourceurl'] is not None:
                if isinstance(self.spec['sourceurl'], str):
                    url = self.spec['sourceurl']
        if url is None:
            raise Exception('No "sourceUrl" found. This field is required.')
        
        target_file: str
        target_file = None
        if 'targetoutputfile' in self.spec:
            if self.spec['targetoutputfile'] is not None:
                if isinstance(self.spec['targetoutputfile'], str):
                    target_file = self.spec['targetoutputfile']
        if target_file is None:
            raise Exception('No "targetOutputFile" found. This field is required.')

        large_file = False
        remote_file_size = self._get_url_content_length(url=self.spec['sourceurl'], log_header=log_header)
        self.log(message='Checking if {} > 104857600...'.format(remote_file_size), build_log_message_header=False, level='info', header=log_header)
        if remote_file_size > 104857600:   # Anything larger than 100MiB is considered large and will be downloaded in chunks
            large_file = True

        # Check if the local file exists:
        if os.path.exists(self.spec['targetoutputfile']) is True:
            if Path(self.spec['targetoutputfile']).is_file() is True:
                local_file_size = int(get_file_size(file_path=self.spec['targetoutputfile']))
                self.log(message='local_file_size={}   remote_file_size={}'.format(local_file_size, remote_file_size), build_log_message_header=False, level='info', header=log_header)
                if local_file_size == remote_file_size:
                    return self._store_values(key_value_store=copy.deepcopy(new_key_value_store), value='n/a', task_id=task.task_id, command=command, context=context, log_header=log_header)
            else:
                raise Exception('The target output file cannot be used as the named target exists but is not a file')

        use_ssl = False
        verify_ssl = True
        use_proxy = False
        use_proxy_authentication = False
        proxy_host = None
        proxy_username = None
        proxy_password = None
        use_http_basic_authentication = False
        http_basic_authentication_username = None
        http_basic_authentication_password = None
        extra_headers = None
        use_custom_headers = False
        http_method = 'GET'
        http_body = None
        use_body = False

        if url.lower().startswith('https'):
            use_ssl = True
        if use_ssl is True and 'skipsslverification' in self.spec:
            if self.spec['skipsslverification'] is not None:
                if isinstance(self.spec['skipsslverification'], bool):
                    verify_ssl = not self.spec['skipsslverification']
                else:
                    self.log(message='Found `skipSslVerification` but value is not a boolean type - ignoring', build_log_message_header=False, level='warning', header=log_header)
            else:
                self.log(message='Found `skipSslVerification` but value is None type - ignoring', build_log_message_header=False, level='warning', header=log_header)

        if 'proxy' in self.spec:
            if self.spec['proxy'] is not None:
                if isinstance(self.spec['proxy'], dict) is True:
                    if 'host' in self.spec['proxy']:
                        if self.spec['proxy']['host'] is not None:
                            if isinstance(self.spec['proxy']['host'], str) is True:
                                use_proxy = True
                                proxy_host = self.spec['proxy']['host']
                                if 'basicauthentication' in self.spec['proxy']:
                                    if self.spec['proxy']['basicauthentication'] is not None:
                                        if isinstance(self.spec['proxy']['basicauthentication'], dict) is True:
                                            use_proxy_authentication = True
                                            if 'username' in ['proxy']['basicauthentication']:
                                                if self.spec['proxy']['basicauthentication']['username'] is not None:
                                                    if isinstance(self.spec['proxy']['basicauthentication']['username'], str) is True:
                                                        proxy_username = self.spec['proxy']['basicauthentication']['username']
                                            if 'password' in ['proxy']['basicauthentication']:
                                                if self.spec['proxy']['basicauthentication']['password'] is not None:
                                                    if isinstance(self.spec['proxy']['basicauthentication']['password'], str) is True:
                                                        proxy_password = self.spec['proxy']['basicauthentication']['password']
                                            if proxy_password is None:
                                                use_proxy_authentication = False

        if 'httpbasicauthentication' in self.spec:
            if self.spec['httpbasicauthentication'] is not None:
                if isinstance(self.spec['httpbasicauthentication'], dict) is True:
                    use_http_basic_authentication = True
                    if 'username' in self.spec['httpbasicauthentication']:
                        if self.spec['httpbasicauthentication']['username'] is not None:
                            if isinstance(self.spec['httpbasicauthentication']['username'], dict) is True:
                                http_basic_authentication_username = self.spec['httpbasicauthentication']['username']
                    if 'password' in self.spec['httpbasicauthentication']:
                        if self.spec['httpbasicauthentication']['password'] is not None:
                            if isinstance(self.spec['httpbasicauthentication']['password'], dict) is True:
                                http_basic_authentication_password =  self.spec['httpbasicauthentication']['password']
                    if http_basic_authentication_password is None or http_basic_authentication_username is None:
                        use_http_basic_authentication = False

        if 'extraheaders' in self.spec:
            if self.spec['extraheaders'] is not None:
                if isinstance(self.spec['extraheaders'], list) is True:
                    extra_headers = dict()
                    for header_data in self.spec['extraheaders']:
                        if header_data is not None:
                            if isinstance(header_data, dict) is True:
                                if 'name' in header_data and 'value' in header_data:
                                    extra_headers[header_data['name']] = header_data['value']
                                else:
                                    self.log(message='      Ignoring extra header item as it does not contain the keys "name" and/or "value"', build_log_message_header=False, level='warning', header=log_header)
        try:
            if len(extra_headers) > 0:
                use_custom_headers = True
        except:
            self.log(message='extra_headers length is zero - not using custom headers', build_log_message_header=False, level='info', header=log_header)

        if 'method' in self.spec:
            if self.spec['method'] is not None:
                if isinstance(self.spec['method'], str) is True:
                    http_method = self.spec['method'].upper()
                    if http_method not in ('GET','HEAD','POST','PUT','DELETE','PATCH',):
                        self.log(message='      HTTP method "{}" not recognized. Defaulting to GET'.format(http_method), build_log_message_header=False, level='warning', header=log_header)
                        http_method = 'GET'

        if http_method != 'GET' and 'body' in self.spec:
            if self.spec['body'] is not None:
                http_body = self.spec['body']
        elif http_method == 'GET' and 'body' in self.spec:
            self.log(message='Body cannot be set with GET requests - ignoring body content', build_log_message_header=False, level='warning', header=log_header)
        if http_body is not None:
            if len(http_body) > 0:
                use_body = True

        self.log(message='   * Large File                      : {}'.format(large_file), build_log_message_header=False, level='info', header=log_header)
        self.log(message='   * Using SSL                       : {}'.format(use_ssl), build_log_message_header=False, level='info', header=log_header)
        if use_ssl:
            self.log(message='   * Skip SSL Verification           : {}'.format(not verify_ssl), build_log_message_header=False, level='info', header=log_header)
        self.log(message='   * Using Proxy                     : {}'.format(use_proxy), build_log_message_header=False, level='info', header=log_header)
        if use_proxy:
            self.log(message='   * Proxy Host                      : {}'.format(proxy_host), build_log_message_header=False, level='info', header=log_header)
            self.log(message='   * Using Proxy Authentication      : {}'.format(use_proxy_authentication), build_log_message_header=False, level='info', header=log_header)
            if use_proxy_authentication is True:
                self.log(message='   * Proxy Password Length           : {}'.format(len(proxy_password)), build_log_message_header=False, level='info', header=log_header)
        self.log(message='   * Using HTTP Basic Authentication : {}'.format(use_http_basic_authentication), build_log_message_header=False, level='info', header=log_header)
        if use_http_basic_authentication:
            self.log(message='   * HTTP Password Length            : {}'.format(len(http_basic_authentication_password)), build_log_message_header=False, level='info', header=log_header)
        if extra_headers is not None:
            if len(extra_headers) > 0:
                self.log(message='   * Extra Header Keys               : {}'.format(list(extra_headers.keys())), build_log_message_header=False, level='info', header=log_header)
            else:
                self.log(message='   * Extra Header Keys               : None - Using Default Headers', build_log_message_header=False, level='info', header=log_header)
        else:
            self.log(message='   * Extra Header Keys               : None - Using Default Headers', build_log_message_header=False, level='info', header=log_header)
        self.log(message='   * HTTP Method                     : {}'.format(http_method), build_log_message_header=False, level='info', header=log_header)
        if http_body is not None:
            self.log(message='   * HTTP Body Bytes                 : {}'.format(len(http_body)), build_log_message_header=False, level='info', header=log_header)
        else:
            self.log(message='   * HTTP Body Bytes                 : None', build_log_message_header=False, level='info', header=log_header)

        work_values = {
            'large_file': large_file,
            'verify_ssl': verify_ssl,
            'use_proxy': use_proxy,
            'use_proxy_authentication': use_proxy_authentication,
            'use_http_basic_authentication': use_http_basic_authentication,
            'http_method': http_method,
            'use_custom_headers': use_custom_headers,
            'use_body': use_body,
        }

        parameters = {
            'url': url,
            'target_file': target_file,
            'verify_ssl': verify_ssl,
            'proxy_host': proxy_host,
            'proxy_username': proxy_username,
            'proxy_password': proxy_password,
            'username': http_basic_authentication_username,
            'password': http_basic_authentication_password,
            'headers': extra_headers,
            'method': http_method,
            'body': http_body
        }
        download_scenarios = [
            {
                'values': {
                    'large_file': False,
                },
                'method': self._get_data_basic_request
            },
            {
                'values': {
                    'large_file': True,
                },
                'method': self._get_data_basic_request_stream
            },
        ]

        effective_method = None
        for scenario in download_scenarios:
            values = scenario['values']
            criterion_match = True
            for criterion, condition in values.items():
                if criterion != 'http_method':
                    if condition != work_values[criterion]:
                        criterion_match = False
                else:
                    if work_values['http_method'] not in condition:
                        criterion_match = False
            if criterion_match is True:
                effective_method = scenario['method']

        if effective_method is not None:
            result = effective_method(**parameters)
            if result is not None:
                new_key_value_store = self._store_values(key_value_store=copy.deepcopy(new_key_value_store), value=result, task_id=task.task_id, command=command, context=context, log_header=log_header)
            else:
                raise Exception('Failed to download "{}" to file "{}"'.format(url, target_file))
        else:
            raise Exception('No suitable method could be found to handle the download.')

        self.spec = dict()
        self.metadata = dict()
        self.log(message='DONE', build_log_message_header=False, level='info', header=log_header)
        return new_key_value_store




import copy
import re
from magnum_opus.operarius import Hook, Task, KeyValueStore, LoggerWrapper, TaskLifecycleStage


def is_iterable(data: object, exclude_dict: bool=True, exclude_string: bool=True)->bool:
    if data is None:
        return False
    if exclude_dict is True and isinstance(data, dict):
        return False
    if exclude_string is True and isinstance(data, str):
        return False
    try:
        iter(data)
    except TypeError as te:
        return False
    return True


def lookup_value(raw_key: str, command:str, context:str, logger:LoggerWrapper, hook_name: str, key_value_store: KeyValueStore)->object:
    # Typical key in key_value_store     : MyKind:MyId:a_command:a_context:SubKey1:SubKey2 (SubKey1 is required, but anything afterwards is optional)
    # Expected raw_key is something like : ${KVS:MyId:SubKey1:SubKey2}  
    result = ''
    logger.debug('[{}]       raw_key: {}'.format(hook_name, raw_key))
    if raw_key.startswith('${KVS:'):                # ${KVS:MyId:SubKey1:SubKey2}        
        key_parts = raw_key.split(':')              # ['${KVS', 'MyId', 'SubKey1', 'SubKey2}']
        target_task_id = key_parts[1]               # MyId
        target_index = ':'.join(key_parts[2:])      # SubKey1:SubKey2
        target_index = target_index.replace('}', '')
        lookup_key_base = '{}:{}:{}:{}'.format(     # MyId:a_command:a_context:SubKey1:SubKey2
            target_task_id,
            command,
            context,
            target_index
        )
        logger.debug('[{}]         Looking for a key that looks like "{}" in key_value_store'.format(hook_name, lookup_key_base))
        # logger.debug('[{}]           Potential keys: {}'.format(hook_name, list(key_value_store.store.keys())))
        for key in list(key_value_store.store.keys()):
            logger.debug('[{}]           Looking for key "{}" in "{}"'.format(hook_name, lookup_key_base, key))
            if lookup_key_base in key:
                logger.info('[{}]             Resolved key "{}" to swap out for reference variable "{}"'.format(hook_name, key, raw_key))
                result = copy.deepcopy(key_value_store.store[key])
    else:
        raise Exception('Oops - the raw key is not what we expected: raw_key: "{}"'.format(raw_key))
    logger.debug('[{}]         Returning final result: "{}"'.format(hook_name, result))
    return result


def analyse_data(data: object, key_value_store:KeyValueStore, command:str, context:str, task_id: str, logger:LoggerWrapper, hook_name: str, task_kind: str)->dict:
    logger.info('[{}]   Analyzing data'.format(hook_name))
    logger.debug('[{}]   task "{}" - Inspecting object: {}'.format(hook_name, task_id, data))
    if data is None:
        return data
    modified_data = None
    if isinstance(data, str) is True:
        modified_data: str = copy.deepcopy(data)
        # matches = re.findall('(\$\{KVS:[\w|\-|\s|:|.|;|_]+\})', 'wc -l ${KVS:prompt_output_path:RESULT} > ${KVS:prompt_output_path:RESULT}_STATS && rm -vf ${KVS:prompt_output_2_path:RESULT}')
        # ['${KVS:prompt_output_path:RESULT}', '${KVS:prompt_output_path:RESULT}', '${KVS:prompt_output_2_path:RESULT}']
        logger.debug('[{}]     Original string: {}'.format(hook_name, data))
        matches = re.findall('(\$\{KVS:[\w|\-|\s|:|.|;|_]+\})', data)
        logger.debug('[{}]       matches: {}'.format(hook_name, matches))
        for match in matches:
            logger.debug('[{}]     Looking up value for variable placeholder "{}"'.format(hook_name, match))
            final_value = lookup_value(
                raw_key=match,
                command=command,
                context=context,
                logger=logger,
                hook_name=hook_name,
                key_value_store=copy.deepcopy(key_value_store)
            )
            modified_data = modified_data.replace(match, final_value)
    elif isinstance(data, dict) is True:
        modified_data: dict = dict()
        for key, val in data.items():
            modified_data[key] = analyse_data(data=val, key_value_store=key_value_store, command=command, context=context, task_id=task_id, logger=logger, hook_name=hook_name, task_kind=task_kind)
    elif is_iterable(data=data) is True:
        modified_data: list = list()
        for val in data:
            modified_data.append(analyse_data(data=val, key_value_store=key_value_store, command=command, context=context, task_id=task_id, logger=logger, hook_name=hook_name, task_kind=task_kind))
    else:
        modified_data = copy.deepcopy(data)

    return modified_data


def spec_variable_key_value_store_resolver(
    hook_name:str,
    task:Task,
    key_value_store:KeyValueStore,
    command:str,
    context:str,
    task_life_cycle_stage: TaskLifecycleStage,
    extra_parameters:dict,
    logger:LoggerWrapper
)->KeyValueStore:
    new_key_value_store = KeyValueStore()
    new_key_value_store.store = copy.deepcopy(key_value_store.store)

    if task_life_cycle_stage is not TaskLifecycleStage.TASK_PRE_PROCESSING_START:
        return new_key_value_store
    
    if 'SpecModifierKey' not in extra_parameters:
        return new_key_value_store
    
    spec_modifier_key = extra_parameters['SpecModifierKey']
    if spec_modifier_key is None:
        return new_key_value_store
    
    if isinstance(spec_modifier_key, str) is False:
        return new_key_value_store
    
    if 'TASK_PRE_PROCESSING_START' not in spec_modifier_key:
        return new_key_value_store

    logger.info('[{}] Called on TASK_PRE_PROCESSING_START hook event for task "{}"'.format(hook_name, task.task_id))
    logger.debug('[{}] spec_modifier_key={}'.format(hook_name, spec_modifier_key))
    logger.debug('[{}]   key_value_store keys: {}'.format(hook_name, list(key_value_store.store.keys())))

    new_key_value_store.save(
        key=spec_modifier_key,
        value=analyse_data(
            data=copy.deepcopy(task.spec),
            key_value_store=copy.deepcopy(key_value_store),
            command=command,
            context=context,
            task_id=task.task_id,
            logger=logger,
            hook_name=hook_name,
            task_kind=task.kind
        )
    )

    return new_key_value_store



import sys
import os
import copy
from inspect import stack
import random
import string
from  pytest_httpserver import HTTPServer
import traceback

sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../src")
print('sys.path={}'.format(sys.path))

import unittest

from opus_instrumenta.task_processors.web_download_file import WebDownloadFile
from opus_instrumenta.task_processors.cli_input_prompt_v1 import CliInputPrompt
from opus_instrumenta.task_processors.shell_script_v1 import ShellScript
from opus_instrumenta.molitor import build_tasks
from magnum_opus.operarius import LoggerWrapper, Task, Tasks, Identifier, Identifiers, IdentifierContext, IdentifierContexts, TaskProcessor, KeyValueStore, build_command_identifier

running_path = os.getcwd()
print('Current Working Path: {}'.format(running_path))


def validate_order(must_be_before_input_task_name: str, input_task_name: str, list_of_tasks: list)->bool:
    must_be_before_input_task_name_pos = list_of_tasks.index(must_be_before_input_task_name)
    input_task_name_pos = list_of_tasks.index(input_task_name)
    return must_be_before_input_task_name_pos < input_task_name_pos


class TestLogger(LoggerWrapper):

    def __init__(self):
        super().__init__()
        self.info_lines = list()
        self.warn_lines = list()
        self.debug_lines = list()
        self.critical_lines = list()
        self.error_lines = list()
        self.all_lines_in_sequence = list()

    def info(self, message: str):
        self.info_lines.append('[LOG] INFO: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.info_lines[-1])
        )

    def warn(self, message: str):
        self.warn_lines.append('[LOG] WARNING: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.warn_lines[-1])
        )

    def warning(self, message: str):
        self.warn_lines.append('[LOG] WARNING: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.warn_lines[-1])
        )

    def debug(self, message: str):
        self.debug_lines.append('[LOG] DEBUG: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.debug_lines[-1])
        )

    def critical(self, message: str):
        self.critical_lines.append('[LOG] CRITICAL: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.critical_lines[-1])
        )

    def error(self, message: str):
        self.error_lines.append('[LOG] ERROR: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.error_lines[-1])
        )

    def reset(self):
        self.info_lines = list()
        self.warn_lines = list()
        self.debug_lines = list()
        self.critical_lines = list()
        self.error_lines = list()


def print_logger_lines(logger:LoggerWrapper):
    for line in logger.all_lines_in_sequence:
        print(line)


def dump_key_value_store(test_class_name: str, test_method_name: str, key_value_store: KeyValueStore):
    try:
        print('\n\n-------------------------------------------------------------------------------')
        print('\t\tTest Class  : {}'.format(test_class_name))
        print('\t\tTest Method : {}'.format(test_method_name))
        print('\n-------------------------------------------------------------------------------')

        # First get the max key length:
        max_key_len = 0
        for key,val in key_value_store.store.items():
            if len(key) > max_key_len:
                max_key_len = len(key)

        for key,val in key_value_store.store.items():
            final_key = '{}'.format(key)
            spaces_qty = max_key_len - len(final_key) + 1
            spaces = ' '*spaces_qty
            final_key = '{}{}: '.format(final_key, spaces)
            print('{}{}\n'.format(final_key, val))

        print('\n_______________________________________________________________________________')
    except:
        pass


class TestScenariosBasicGet(unittest.TestCase):    # pragma: no cover

    def setUp(self) -> None:
        print()
        print('-'*80)
        self.logger = TestLogger()
        return super().setUp()

    def tearDown(self):
        print_logger_lines(logger=self.logger)
        self.logger = None
        return super().tearDown()

    def test_calculated_task_processing_order_01(self):
        task_prompt_for_source_url = Task(
            kind='CliInputPrompt',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "prompt_url"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ]
            },
            spec={
                'promptText': 'Get a web URL for a file to download or use a default value after the timeout.',
                'defaultValue': 'https://raw.githubusercontent.com/nicc777/py-animus-extensions/main/implementations/web-download-file-v1.py',
                'promptCharacter': 'URL:',
                'waitTimeoutSeconds': 5,
            },
            logger=self.logger
        )
        task_prompt_for_output_file = Task(
            kind='CliInputPrompt',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "prompt_output_path"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ]
            },
            spec={
                'promptText': 'Supply a destination file where you want to save this file.',
                'defaultValue': '/tmp/output',
                'promptCharacter': 'PATH:',
                'waitTimeoutSeconds': 5,
            },
            logger=self.logger
        )
        task_download = Task(
            kind='WebDownloadFile',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "download"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ],
                "dependencies": [
                    {
                        "identifierType": "ManifestName",
                        "identifiers": [
                            { "key": "prompt_url" },
                            { "key": "prompt_output_path" },
                        ]
                    }
                ]
            },
            spec={
                'sourceUrl': '${KVS:prompt_url:RESULT}',
                'targetOutputFile': '${KVS:prompt_output_path:RESULT}',
            },
            logger=self.logger
        )
        task_file_test_and_cleanup = Task(
            kind='ShellScript',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "stats_and_cleanup"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ],
                "dependencies": [
                    {
                        "identifierType": "ManifestName",
                        "identifiers": [
                            { "key": "download" },
                        ]
                    }
                ]
            },
            spec={
                'source': {
                    'type': 'inline',
                    'value': 'wc -l ${KVS:prompt_output_path:RESULT} > ${KVS:prompt_output_path:RESULT}_STATS && rm -vf ${KVS:prompt_output_path:RESULT}'
                }
            },
            logger=self.logger
        )
        tasks = Tasks(logger=self.logger)

        cli_input_processor = CliInputPrompt(logger=self.logger)
        web_download_processor = WebDownloadFile(logger=self.logger)
        shell_script_processor = ShellScript(logger=self.logger)

        tasks.register_task_processor(processor=cli_input_processor)
        tasks.register_task_processor(processor=web_download_processor)
        tasks.register_task_processor(processor=shell_script_processor)

        tasks.add_task(task=task_prompt_for_source_url)
        tasks.add_task(task=task_file_test_and_cleanup)
        tasks.add_task(task=task_download)
        tasks.add_task(task=task_prompt_for_output_file)

        processing_target_identifier = build_command_identifier(command='apply', context='unittest')
        calculated_task_order = tasks.calculate_current_task_order(processing_target_identifier=processing_target_identifier)
        print('***   calculated_task_order={}'.format(calculated_task_order))
        self.assertTrue(validate_order(must_be_before_input_task_name='prompt_url', input_task_name='prompt_output_path', list_of_tasks=calculated_task_order), 'FAILED TEST: Expected prompt_url to be before prompt_output_path')
        self.assertTrue(validate_order(must_be_before_input_task_name='prompt_url', input_task_name='download', list_of_tasks=calculated_task_order), 'FAILED TEST: Expected prompt_url to be before download')
        self.assertTrue(validate_order(must_be_before_input_task_name='prompt_url', input_task_name='stats_and_cleanup', list_of_tasks=calculated_task_order), 'FAILED TEST: Expected prompt_url to be before stats_and_cleanup')
        self.assertTrue(validate_order(must_be_before_input_task_name='prompt_output_path', input_task_name='download', list_of_tasks=calculated_task_order), 'FAILED TEST: Expected prompt_output_path to be before download')
        self.assertTrue(validate_order(must_be_before_input_task_name='prompt_output_path', input_task_name='stats_and_cleanup', list_of_tasks=calculated_task_order), 'FAILED TEST: Expected prompt_output_path to be before stats_and_cleanup')
        self.assertTrue(validate_order(must_be_before_input_task_name='download', input_task_name='stats_and_cleanup', list_of_tasks=calculated_task_order), 'FAILED TEST: Expected download to be before stats_and_cleanup')

    def test_processing_using_hooks_for_variable_substitution_01(self):
        task_prompt_for_source_url = Task(
            kind='CliInputPrompt',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "prompt_url"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ]
            },
            spec={
                'promptText': 'Get a web URL for a file to download or use a default value after the timeout.',
                'defaultValue': 'https://raw.githubusercontent.com/nicc777/py-animus-extensions/main/implementations/web-download-file-v1.py',
                'promptCharacter': 'URL:',
                'waitTimeoutSeconds': 5,
            },
            logger=self.logger
        )
        task_prompt_for_output_file = Task(
            kind='CliInputPrompt',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "prompt_output_path"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ]
            },
            spec={
                'promptText': 'Supply a destination file where you want to save this file.',
                'defaultValue': '/tmp/output',
                'promptCharacter': 'PATH:',
                'waitTimeoutSeconds': 5,
            },
            logger=self.logger
        )
        task_download = Task(
            kind='WebDownloadFile',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "download"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ],
                "dependencies": [
                    {
                        "identifierType": "ManifestName",
                        "identifiers": [
                            { "key": "prompt_url" },
                            { "key": "prompt_output_path" },
                        ]
                    }
                ]
            },
            spec={
                'sourceUrl': '${KVS:prompt_url:RESULT}',
                'targetOutputFile': '${KVS:prompt_output_path:RESULT}',
            },
            logger=self.logger
        )
        task_file_test_and_cleanup = Task(
            kind='ShellScript',
            version='v1',
            metadata={
                "identifiers": [
                    {
                        "type": "ManifestName",
                        "key": "stats_and_cleanup"
                    },
                    {
                        "type": "Label",
                        "key": "is_unittest",
                        "value": "TRUE"
                    }
                ],
                "dependencies": [
                    {
                        "identifierType": "ManifestName",
                        "identifiers": [
                            { "key": "download" },
                        ]
                    }
                ]
            },
            spec={
                'source': {
                    'type': 'inline',
                    'value': 'wc -l ${KVS:prompt_output_path:RESULT} > ${KVS:prompt_output_path:RESULT}_STATS && rm -vf ${KVS:prompt_output_path:RESULT}'
                }
            },
            logger=self.logger
        )

        tasks = build_tasks(logger=self.logger)

        tasks.add_task(task=task_prompt_for_source_url)
        tasks.add_task(task=task_file_test_and_cleanup)
        tasks.add_task(task=task_download)
        tasks.add_task(task=task_prompt_for_output_file)

        processing_target_identifier = build_command_identifier(command='apply', context='unittest')
        calculated_task_order = tasks.calculate_current_task_order(processing_target_identifier=processing_target_identifier)
        print('***   calculated_task_order={}'.format(calculated_task_order))
        self.assertTrue(validate_order(must_be_before_input_task_name='prompt_url', input_task_name='prompt_output_path', list_of_tasks=calculated_task_order), 'FAILED TEST: Expected prompt_url to be before prompt_output_path')
        self.assertTrue(validate_order(must_be_before_input_task_name='prompt_url', input_task_name='download', list_of_tasks=calculated_task_order), 'FAILED TEST: Expected prompt_url to be before download')
        self.assertTrue(validate_order(must_be_before_input_task_name='prompt_url', input_task_name='stats_and_cleanup', list_of_tasks=calculated_task_order), 'FAILED TEST: Expected prompt_url to be before stats_and_cleanup')
        self.assertTrue(validate_order(must_be_before_input_task_name='prompt_output_path', input_task_name='download', list_of_tasks=calculated_task_order), 'FAILED TEST: Expected prompt_output_path to be before download')
        self.assertTrue(validate_order(must_be_before_input_task_name='prompt_output_path', input_task_name='stats_and_cleanup', list_of_tasks=calculated_task_order), 'FAILED TEST: Expected prompt_output_path to be before stats_and_cleanup')
        self.assertTrue(validate_order(must_be_before_input_task_name='download', input_task_name='stats_and_cleanup', list_of_tasks=calculated_task_order), 'FAILED TEST: Expected download to be before stats_and_cleanup')

        try:
            tasks.process_context(command='test', context='unittest')
        except:
            # traceback.format_exc()
            print_logger_lines(logger=self.logger)
            self.fail('Task processing caught an exception. Please review the logs.')

        # TODO Test if file '/tmp/output_STATS' exists. Must have content similar to: "393 /tmp/output" (INTEGER>0, SPACE, String "/tmp/output")
        # TODO Delete '/tmp/output_STATS'


if __name__ == '__main__':
    unittest.main()



import json
import traceback
from getpass import getpass
import copy
import signal
import contextlib

from magnum_opus.operarius import LoggerWrapper, TaskProcessor, KeyValueStore, Task, StatePersistence


class TimeoutException(Exception):
    pass


@contextlib.contextmanager
def timeout_context(seconds):
    def signal_handler(signum, frame):
        raise TimeoutException("Timed out!")
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


def get_normal_user_input_with_timeout(prompt_char: str='> ', timeout_seconds: int=10, default_value: str=''):
    result = default_value
    try:
        with timeout_context(timeout_seconds):
            result = input(prompt_char)
            if result == '':
                result = default_value
    except TimeoutException:
        print(default_value)
    return result


def get_password_input_with_timeout(prompt_char: str='> ', timeout_seconds: int=10, default_value: str=''):
    result = default_value
    try:
        with timeout_context(timeout_seconds):
            result = getpass(prompt=prompt_char)
            if result == '':
                result = default_value
    except TimeoutException:
        print()
    return result


class CliInputPrompt(TaskProcessor):

    def __init__(self, kind: str='CliInputPrompt', kind_versions: list=['v1',], supported_commands: list = list(), logger: LoggerWrapper = LoggerWrapper()):
        self.spec = dict()
        self.metadata = dict()
        super().__init__(kind, kind_versions, supported_commands, logger)

    def process_task(self, task: Task, command: str, context: str = 'default', key_value_store: KeyValueStore = KeyValueStore(), state_persistence: StatePersistence = StatePersistence()) -> KeyValueStore:
        """This task will prompt a user for input. If no input is provided with an (optional) timeout, a default value 
        will be used. The user input value will be added to the `KeyValueStore`.
        
        Regardless of command and context, the specified shell script will be run, unless specifically excluded.

        # Spec fields

        Root levels spec fields

        | Field                         | Type    | Required | In Versions | Description                                                                                                                                                                                                          |
        |-------------------------------|:-------:|:--------:|:-----------:|----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
        | `promptText`                  | String  | No       | v1          | The text to display on screen. If set, must be a string with minimum length of 2 and less then 80 characters.                                                                                                        |
        | `defaultValue`                | String  | No       | v1          | If defined, this will be the default value a user can choose by just pressing ENTER, or it will also be the value used after the `waitTimeout` expires.                                                              |
        | `promptCharacter`             | String  | No       | v1          | The character for the actual prompt. If set, must be a string with minimum length of 1 and less then 8 characters.                                                                                                   |
        | `waitTimeoutSeconds`          | Integer | No       | v1          | Default is "0" (do not expire - wait forever). Any value >0 but <3600 (1 hour) will wait for user input. When the timeout is reached, the default value will be used.                                                |
        | `convertEmptyInputToNone`     | Boolean | No       | v1          | If input is empty, convert the final value to NoneType                                                                                                                                                               |
        | `maskInput`                   | Boolean | No       | v1          | If true, do not echo characters. This is suitable to ask for a password, for example                                                                                                                                 |


        Args:
            task: The `Task` of kind `ShellScript` version `v1` to process
            command: The command is ignored for ShellScript processing - any task of this kind will ALWAYS be processed regardless of the command.
            context: The context is ignored for ShellScript processing - any task of this kind will ALWAYS be processed regardless of the context.
            key_value_store: An instance of the `KeyValueStore`. If none is supplied, a new instance will be created.
            state_persistence: An implementation of `StatePersistence` that the task processor can use to retrieve previous copies of the `Task` manifest in order to determine the actions to be performed.

        Returns:
            An updated `KeyValueStore`.

            * Value from the user's input, or the default value if the `waitTimeout` was reached: `task.kind:task.task_id:command:context:RESULT`

        Raises:
            Exception: As determined by the processing logic.
        """
        self.spec = copy.deepcopy(task.spec)
        self.metadata = copy.deepcopy(task.metadata)
        new_key_value_store = KeyValueStore()
        new_key_value_store.store = copy.deepcopy(key_value_store.store)
        log_header = self.format_log_header(task=task, command=command, context=context)
        self.log(message='PROCESSING START', build_log_message_header=False, level='info', header=log_header)
        self.log(message='   spec: {}'.format(json.dumps(self.spec)), build_log_message_header=False, level='debug', header=log_header)
        if '{}:{}:{}:{}:RESULT'.format(task.kind,task.task_id,command,context) in key_value_store.store is True:
            self.log(message='The task have already been processed and will now be ignored. The KeyValueStore will be returned unmodified.', build_log_message_header=False, level='warning', header=log_header)
            return new_key_value_store

        mask_input = False
        default_value = ''
        wait_timeout_seconds = 0
        prompt_text = None
        prompt_char = '> '
        convert_empty_input_to_none_value = False
        # IMPORTANT: Remember that all keys were converted to LOWERCASE
        if 'prompttext' in self.spec:
            if self.spec['prompttext'] is not None:
                if isinstance(self.spec['prompttext'], str):
                    if len(self.spec['prompttext']) > 1 and len(self.spec['prompttext']) < 80:
                        prompt_text = self.spec['prompttext']
        if 'promptcharacter' in self.spec:
            if self.spec['promptcharacter'] is not None:
                if isinstance(self.spec['promptcharacter'], str):
                    if len(self.spec['promptcharacter']) > 0 and len(self.spec['promptcharacter']) < 8:
                        prompt_char = '{} '.format(self.spec['promptcharacter'])
        if 'defaultvalue' in self.spec:
            if self.spec['defaultvalue'] is not None:
                if isinstance(self.spec['defaultvalue'], str):
                    if len(self.spec['defaultvalue']) > 1 and len(self.spec['defaultvalue']) < 256:
                        default_value = self.spec['defaultvalue']
                        prompt_char = '[default={}] {}'.format(default_value, prompt_char)
        if 'maskinput' in self.spec:
            if self.spec['maskinput'] is not None:
                if isinstance(self.spec['maskinput'], bool):
                    mask_input = self.spec['maskinput']
        if 'waittimeoutseconds' in self.spec:
            if self.spec['waittimeoutseconds'] is not None:
                if isinstance(self.spec['waittimeoutseconds'], int):
                    if self.spec['waittimeoutseconds'] > 0 and self.spec['waittimeoutseconds'] < 3600:
                        wait_timeout_seconds = self.spec['waittimeoutseconds']
        if 'convertemptyinputtonone' in self.spec:
            if self.spec['convertemptyinputtonone'] is not None:
                if isinstance(self.spec['convertemptyinputtonone'], bool):
                    convert_empty_input_to_none_value = self.spec['convertemptyinputtonone']
        
        """
            mask_input = False
            default_value = ''
            wait_timeout_seconds = 0
            prompt_text = None
            prompt_char = '> '
            convert_empty_input_to_none_value = False
        """
        self.log(message='FINAL VALUE for mask_input: "{}" (type={})'.format(mask_input, type(mask_input)), build_log_message_header=False, level='debug', header=log_header)
        self.log(message='FINAL VALUE for default_value: "{}" (type={})'.format(default_value, type(default_value)), build_log_message_header=False, level='debug', header=log_header)
        self.log(message='FINAL VALUE for wait_timeout_seconds: "{}" (type={})'.format(wait_timeout_seconds, type(wait_timeout_seconds)), build_log_message_header=False, level='debug', header=log_header)
        self.log(message='FINAL VALUE for prompt_text: "{}" (type={})'.format(prompt_text, type(prompt_text)), build_log_message_header=False, level='debug', header=log_header)
        self.log(message='FINAL VALUE for prompt_char: "{}" (type={})'.format(prompt_char, type(prompt_char)), build_log_message_header=False, level='debug', header=log_header)
        self.log(message='FINAL VALUE for convert_empty_input_to_none_value: "{}" (type={})'.format(convert_empty_input_to_none_value, type(convert_empty_input_to_none_value)), build_log_message_header=False, level='debug', header=log_header)
        
        if prompt_text is not None:
            print('{}\n'.format(prompt_text))

        value = None
        try:
            if wait_timeout_seconds > 0:
                if mask_input is True:
                    value = get_password_input_with_timeout(prompt_char=prompt_char, timeout_seconds=wait_timeout_seconds, default_value=default_value)
                else:
                    value = get_normal_user_input_with_timeout(prompt_char=prompt_char, timeout_seconds=wait_timeout_seconds, default_value=default_value)
            else:
                if mask_input is True:
                    value = getpass(prompt=prompt_char)
                    if value == '':
                        value = default_value
                else:
                    value = input(prompt_char)
                    if value == '':
                        value = default_value
        except:
            self.log(message='EXCEPTION: {}'.format(traceback.format_exc()), build_log_message_header=False, level='error', header=log_header)
            self.log(message='Using DEFAULT value', build_log_message_header=False, level='warning', header=log_header)
            value = default_value
        self.log(message='  Storing value: "{}"'.format(value), build_log_message_header=False, level='info', header=log_header)
        new_key_value_store.save(key='{}:{}:{}:{}:RESULT'.format(task.kind, task.task_id, command, context), value=value)

        self.log(message='value={}'.format(value), build_log_message_header=False, level='debug', header=log_header)
        if value == '' and convert_empty_input_to_none_value is True:
            value = None
        self.log(message='final value={}'.format(value), build_log_message_header=False, level='debug', header=log_header)

        self.spec = dict()
        self.metadata = dict()
        self.log(message='DONE', build_log_message_header=False, level='info', header=log_header)
        self.log(message='  key_value_store keys: {}'.format(list(new_key_value_store.store.keys())), level='debug', build_log_message_header=False, header=log_header)
        return new_key_value_store




import sys
import os
import copy
from inspect import stack
import random
import string
from  pytest_httpserver import HTTPServer

sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../src")
print('sys.path={}'.format(sys.path))

import unittest

from opus_instrumenta.task_processors.web_download_file import WebDownloadFile
from magnum_opus.operarius import LoggerWrapper, Task, Tasks, Identifier, Identifiers, IdentifierContext, IdentifierContexts, TaskProcessor, KeyValueStore

running_path = os.getcwd()
print('Current Working Path: {}'.format(running_path))


class TestLogger(LoggerWrapper):

    def __init__(self):
        super().__init__()
        self.info_lines = list()
        self.warn_lines = list()
        self.debug_lines = list()
        self.critical_lines = list()
        self.error_lines = list()
        self.all_lines_in_sequence = list()

    def info(self, message: str):
        self.info_lines.append('[LOG] INFO: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.info_lines[-1])
        )

    def warn(self, message: str):
        self.warn_lines.append('[LOG] WARNING: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.warn_lines[-1])
        )

    def warning(self, message: str):
        self.warn_lines.append('[LOG] WARNING: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.warn_lines[-1])
        )

    def debug(self, message: str):
        self.debug_lines.append('[LOG] DEBUG: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.debug_lines[-1])
        )

    def critical(self, message: str):
        self.critical_lines.append('[LOG] CRITICAL: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.critical_lines[-1])
        )

    def error(self, message: str):
        self.error_lines.append('[LOG] ERROR: {}'.format(message))
        self.all_lines_in_sequence.append(
            copy.deepcopy(self.error_lines[-1])
        )

    def reset(self):
        self.info_lines = list()
        self.warn_lines = list()
        self.debug_lines = list()
        self.critical_lines = list()
        self.error_lines = list()


def print_logger_lines(logger:LoggerWrapper):
    for line in logger.all_lines_in_sequence:
        print(line)


def dump_key_value_store(test_class_name: str, test_method_name: str, key_value_store: KeyValueStore):
    try:
        print('\n\n-------------------------------------------------------------------------------')
        print('\t\tTest Class  : {}'.format(test_class_name))
        print('\t\tTest Method : {}'.format(test_method_name))
        print('\n-------------------------------------------------------------------------------')

        # First get the max key length:
        max_key_len = 0
        for key,val in key_value_store.store.items():
            if len(key) > max_key_len:
                max_key_len = len(key)

        for key,val in key_value_store.store.items():
            final_key = '{}'.format(key)
            spaces_qty = max_key_len - len(final_key) + 1
            spaces = ' '*spaces_qty
            final_key = '{}{}: '.format(final_key, spaces)
            print('{}{}\n'.format(final_key, val))

        print('\n_______________________________________________________________________________')
    except:
        pass


class TestScenariosBasicGet(unittest.TestCase):    # pragma: no cover

    def setUp(self) -> None:
        print()
        print('-'*80)
        self.logger = TestLogger()
        return super().setUp()

    def tearDown(self):
        print_logger_lines(logger=self.logger)
        self.logger = None
        return super().tearDown()

    def test_get_file_01(self):
        test_method_name = '{}'.format(stack()[0][3])
        expected_status_code = 200
        # Generate some random data
        random_data = ''.join(random.choice(string.ascii_letters) for _ in range(100))

        with HTTPServer() as httpserver:
            httpserver.expect_request("/some_file.txt").respond_with_data(response_data=random_data)

            # Process
            web_download_processor = WebDownloadFile(logger=self.logger)
            task = Task(
                kind='WebDownloadFile',
                version='v1',
                metadata={
                    "identifiers": [
                        {
                            "type": "ManifestName",
                            "key": "{}".format(test_method_name)
                        },
                        {
                            "type": "Label",
                            "key": "is_unittest",
                            "value": "TRUE"
                        }
                    ]
                },
                spec={
                    'sourceUrl': httpserver.url_for("/some_file.txt"),
                    'targetOutputFile': '/tmp/output_{}.txt'.format(test_method_name),
                },
                logger=self.logger
            )
            tasks = Tasks(logger=self.logger)
            tasks.register_task_processor(processor=web_download_processor)
            tasks.add_task(task=task)
            tasks.process_context(command='apply', context='unittest')
            tasks.state_persistence.persist_all_state()
            dump_key_value_store(test_class_name=self.__class__.__name__, test_method_name=test_method_name, key_value_store=tasks.key_value_store)
            self.assertIsNotNone(tasks.key_value_store)
            self.assertIsNotNone(tasks.key_value_store.store)
            self.assertIsInstance(tasks.key_value_store, KeyValueStore)
            self.assertIsInstance(tasks.key_value_store.store, dict)
            self.assertTrue('PROCESSING_TASK:{}:apply:unittest'.format(test_method_name) in tasks.key_value_store.store)
            self.assertTrue('WebDownloadFile:{}:apply:unittest:RESULT'.format(test_method_name) in tasks.key_value_store.store)
            self.assertEqual(tasks.key_value_store.store['WebDownloadFile:{}:apply:unittest:RESULT'.format(test_method_name)], '{}'.format(expected_status_code))

        os.unlink('/tmp/output_{}.txt'.format(test_method_name))


if __name__ == '__main__':
    unittest.main()
